@@ -37,6 +37,8 @@
 #include "gproxyvolume.h"
 #include "gproxymount.h"
 
+#include "gproxymountoperation.h"
+
 static void signal_emit_in_idle (gpointer object, const char *signal_name, gpointer other_object);
 
 /* Protects all fields of GProxyVolume that can change */
@@ -63,22 +65,13 @@ struct _GProxyVolume {
   gboolean should_automount;
 
   GProxyShadowMount *shadow_mount;
-
-  GHashTable *hash_mount_op_id_to_data;
 };
 
 static void g_proxy_volume_volume_iface_init (GVolumeIface *iface);
 
-#define _G_IMPLEMENT_INTERFACE_DYNAMIC(TYPE_IFACE, iface_init)       { \
-  const GInterfaceInfo g_implement_interface_info = { \
-    (GInterfaceInitFunc) iface_init, NULL, NULL \
-  }; \
-  g_type_module_add_interface (type_module, g_define_type_id, TYPE_IFACE, &g_implement_interface_info); \
-}
-
 G_DEFINE_DYNAMIC_TYPE_EXTENDED (GProxyVolume, g_proxy_volume, G_TYPE_OBJECT, 0,
-                                _G_IMPLEMENT_INTERFACE_DYNAMIC (G_TYPE_VOLUME,
-                                                                g_proxy_volume_volume_iface_init))
+                                G_IMPLEMENT_INTERFACE_DYNAMIC (G_TYPE_VOLUME,
+                                                               g_proxy_volume_volume_iface_init))
 
 
 static void union_monitor_mount_added (GVolumeMonitor *union_monitor,
@@ -127,8 +120,6 @@ g_proxy_volume_finalize (GObject *object)
       g_object_unref (volume->volume_monitor);
     }
 
-  g_hash_table_unref (volume->hash_mount_op_id_to_data);
-
   if (G_OBJECT_CLASS (g_proxy_volume_parent_class)->finalize)
     (*G_OBJECT_CLASS (g_proxy_volume_parent_class)->finalize) (object);
 }
@@ -185,7 +176,6 @@ g_proxy_volume_class_finalize (GProxyVolumeClass *klass)
 static void
 g_proxy_volume_init (GProxyVolume *proxy_volume)
 {
-  proxy_volume->hash_mount_op_id_to_data = g_hash_table_new (g_str_hash, g_str_equal);
 }
 
 GProxyVolume *
@@ -598,11 +588,12 @@ eject_wrapper_callback (GObject *source_object,
 }
 
 static void
-g_proxy_volume_eject (GVolume              *volume,
-                    GMountUnmountFlags   flags,
-                    GCancellable        *cancellable,
-                    GAsyncReadyCallback  callback,
-                    gpointer             user_data)
+g_proxy_volume_eject_with_operation (GVolume             *volume,
+                                     GMountUnmountFlags   flags,
+                                     GMountOperation     *mount_operation,
+                                     GCancellable        *cancellable,
+                                     GAsyncReadyCallback  callback,
+                                     gpointer             user_data)
 {
   GProxyVolume *proxy_volume = G_PROXY_VOLUME (volume);
   GProxyDrive *drive;
@@ -623,15 +614,15 @@ g_proxy_volume_eject (GVolume              *volume,
       data->object = g_object_ref (volume);
       data->callback = callback;
       data->user_data = user_data;
-      g_drive_eject (G_DRIVE (drive), flags, cancellable, eject_wrapper_callback, data);
+      g_drive_eject_with_operation (G_DRIVE (drive), flags, mount_operation, cancellable, eject_wrapper_callback, data);
       g_object_unref (drive);
     }
 }
 
 static gboolean
-g_proxy_volume_eject_finish (GVolume        *volume,
-                          GAsyncResult  *result,
-                          GError       **error)
+g_proxy_volume_eject_with_operation_finish (GVolume        *volume,
+                                            GAsyncResult  *result,
+                                            GError       **error)
 {
   GProxyVolume *proxy_volume = G_PROXY_VOLUME (volume);
   GProxyDrive *drive;
@@ -647,13 +638,35 @@ g_proxy_volume_eject_finish (GVolume        *volume,
 
   if (drive != NULL)
     {
-      res = g_drive_eject_finish (G_DRIVE (drive), result, error);
+      res = g_drive_eject_with_operation_finish (G_DRIVE (drive), result, error);
       g_object_unref (drive);
     }
 
   return res;
 }
 
+/* ---------------------------------------------------------------------------------------------------- */
+
+static void
+g_proxy_volume_eject (GVolume              *volume,
+                      GMountUnmountFlags   flags,
+                      GCancellable        *cancellable,
+                      GAsyncReadyCallback  callback,
+                      gpointer             user_data)
+{
+  g_proxy_volume_eject_with_operation (volume, flags, NULL, cancellable, callback, user_data);
+}
+
+static gboolean
+g_proxy_volume_eject_finish (GVolume        *volume,
+                             GAsyncResult  *result,
+                             GError       **error)
+{
+  return g_proxy_volume_eject_with_operation_finish (volume, result, error);
+}
+
+/* ---------------------------------------------------------------------------------------------------- */
+
 static char *
 g_proxy_volume_get_identifier (GVolume              *volume,
                                const char          *kind)
@@ -705,9 +718,7 @@ typedef struct {
   GCancellable *cancellable;
   gulong cancelled_handler_id;
 
-  gchar *mount_op_id;
-  GMountOperation *mount_operation;
-  gulong reply_handler_id;
+  const gchar *mount_op_id;
 } DBusOp;
 
 static void
@@ -737,16 +748,9 @@ mount_cb (DBusMessage *reply,
     }
 
   /* free DBusOp */
-  if (strlen (data->mount_op_id) > 0)
-    g_hash_table_remove (data->volume->hash_mount_op_id_to_data, data->mount_op_id);
+  g_proxy_mount_operation_destroy (data->mount_op_id);
   g_object_unref (data->volume);
 
-  g_free (data->mount_op_id);
-  if (data->reply_handler_id > 0)
-    g_signal_handler_disconnect (data->mount_operation, data->reply_handler_id);
-  if (data->mount_operation != NULL)
-    g_object_unref (data->mount_operation);
-
   g_free (data->cancellation_id);
   if (data->cancellable != NULL)
     g_object_unref (data->cancellable);
@@ -902,18 +906,7 @@ g_proxy_volume_mount (GVolume             *volume,
           data->cancellation_id = g_strdup ("");
         }
 
-      if (mount_operation != NULL)
-        {
-          data->mount_op_id = g_strdup_printf ("%p", mount_operation);
-          data->mount_operation = g_object_ref (mount_operation);
-          g_hash_table_insert (proxy_volume->hash_mount_op_id_to_data,
-                               data->mount_op_id,
-                               data);
-        }
-      else
-        {
-          data->mount_op_id = g_strdup ("");
-        }
+      data->mount_op_id = g_proxy_mount_operation_wrap (mount_operation, proxy_volume->volume_monitor);
 
       connection = g_proxy_volume_monitor_get_dbus_connection (proxy_volume->volume_monitor);
       name = g_proxy_volume_monitor_get_dbus_name (proxy_volume->volume_monitor);
@@ -936,7 +929,7 @@ g_proxy_volume_mount (GVolume             *volume,
 
       _g_dbus_connection_call_async (connection,
                                      message,
-                                     30 * 60 * 1000,                /* 30 minute timeout */
+                                     G_PROXY_VOLUME_MONITOR_DBUS_TIMEOUT, /* 30 minute timeout */
                                      (GAsyncDBusCallback) mount_cb,
                                      data);
       dbus_message_unref (message);
@@ -947,225 +940,6 @@ g_proxy_volume_mount (GVolume             *volume,
   ;
 }
 
-
-static void
-mount_op_reply_cb (DBusMessage *reply,
-                   GError      *error,
-                   DBusOp      *data)
-{
-  if (error != NULL)
-    {
-      g_warning ("Error from MountOpReply(): %s", error->message);
-    }
-}
-
-static void
-mount_operation_reply (GMountOperation        *mount_operation,
-                       GMountOperationResult  result,
-                       gpointer               user_data)
-{
-  DBusOp *data = user_data;
-  DBusConnection *connection;
-  const char *name;
-  DBusMessage *message;
-  const char *user_name;
-  const char *domain;
-  const char *password;
-  char *encoded_password;
-  dbus_uint32_t password_save;
-  dbus_uint32_t choice;
-  dbus_bool_t anonymous;
-
-  connection = g_proxy_volume_monitor_get_dbus_connection (data->volume->volume_monitor);
-  name = g_proxy_volume_monitor_get_dbus_name (data->volume->volume_monitor);
-
-  user_name     = g_mount_operation_get_username (mount_operation);
-  domain        = g_mount_operation_get_domain (mount_operation);
-  password      = g_mount_operation_get_password (mount_operation);
-  password_save = g_mount_operation_get_password_save (mount_operation);
-  choice        = g_mount_operation_get_choice (mount_operation);
-  anonymous     = g_mount_operation_get_anonymous (mount_operation);
-
-  if (user_name == NULL)
-    user_name = "";
-  if (domain == NULL)
-    domain = "";
-  if (password == NULL)
-    password = "";
-
-  /* NOTE: this is not to add "security", it's merely to prevent accidental exposure
-   *       of passwords when running dbus-monitor
-   */
-  encoded_password = g_base64_encode ((const guchar *) password, (gsize) (strlen (password) + 1));
-
-  message = dbus_message_new_method_call (name,
-                                          "/org/gtk/Private/RemoteVolumeMonitor",
-                                          "org.gtk.Private.RemoteVolumeMonitor",
-                                          "MountOpReply");
-  dbus_message_append_args (message,
-                            DBUS_TYPE_STRING,
-                            &(data->volume->id),
-                            DBUS_TYPE_STRING,
-                            &(data->mount_op_id),
-                            DBUS_TYPE_INT32,
-                            &result,
-                            DBUS_TYPE_STRING,
-                            &user_name,
-                            DBUS_TYPE_STRING,
-                            &domain,
-                            DBUS_TYPE_STRING,
-                            &encoded_password,
-                            DBUS_TYPE_INT32,
-                            &password_save,
-                            DBUS_TYPE_INT32,
-                            &choice,
-                            DBUS_TYPE_BOOLEAN,
-                            &anonymous,
-                            DBUS_TYPE_INVALID);
-
-  _g_dbus_connection_call_async (connection,
-                                 message,
-                                 -1,
-                                 (GAsyncDBusCallback) mount_op_reply_cb,
-                                 data);
-
-  g_free (encoded_password);
-  dbus_message_unref (message);
-  dbus_connection_unref (connection);
-}
-
-void
-g_proxy_volume_handle_mount_op_ask_password (GProxyVolume        *volume,
-                                             DBusMessageIter     *iter)
-{
-  const char *mount_op_id;
-  const char *message;
-  const char *default_user;
-  const char *default_domain;
-  dbus_int32_t flags;
-  DBusOp *data;
-
-  dbus_message_iter_get_basic (iter, &mount_op_id);
-  dbus_message_iter_next (iter);
-
-  dbus_message_iter_get_basic (iter, &message);
-  dbus_message_iter_next (iter);
-
-  dbus_message_iter_get_basic (iter, &default_user);
-  dbus_message_iter_next (iter);
-
-  dbus_message_iter_get_basic (iter, &default_domain);
-  dbus_message_iter_next (iter);
-
-  dbus_message_iter_get_basic (iter, &flags);
-  dbus_message_iter_next (iter);
-
-  data = g_hash_table_lookup (volume->hash_mount_op_id_to_data, mount_op_id);
-
-  /* since eavesdropping is enabled on the session bus we get this signal even if it
-   * is for another application; so silently ignore it if it's not for us
-   */
-  if (data == NULL)
-    goto out;
-
-  if (data->reply_handler_id == 0)
-    {
-      data->reply_handler_id = g_signal_connect (data->mount_operation,
-                                                 "reply",
-                                                 G_CALLBACK (mount_operation_reply),
-                                                 data);
-    }
-
-  g_signal_emit_by_name (data->mount_operation,
-                         "ask-password",
-                         message,
-                         default_user,
-                         default_domain,
-                         flags);
-
- out:
-  ;
-}
-
-void
-g_proxy_volume_handle_mount_op_ask_question (GProxyVolume        *volume,
-                                             DBusMessageIter     *iter)
-{
-  const char *mount_op_id;
-  const char *message;
-  GPtrArray *choices;
-  DBusMessageIter iter_array;
-  DBusOp *data;
-
-  choices = NULL;
-
-  dbus_message_iter_get_basic (iter, &mount_op_id);
-  dbus_message_iter_next (iter);
-
-  dbus_message_iter_get_basic (iter, &message);
-  dbus_message_iter_next (iter);
-
-  choices = g_ptr_array_new ();
-  dbus_message_iter_recurse (iter, &iter_array);
-  while (dbus_message_iter_get_arg_type (&iter_array) != DBUS_TYPE_INVALID)
-    {
-      const char *choice;
-      dbus_message_iter_get_basic (&iter_array, &choice);
-      dbus_message_iter_next (&iter_array);
-
-      g_ptr_array_add (choices, g_strdup (choice));
-    }
-  g_ptr_array_add (choices, NULL);
-
-  data = g_hash_table_lookup (volume->hash_mount_op_id_to_data, mount_op_id);
-
-  /* since eavesdropping is enabled on the session bus we get this signal even if it
-   * is for another application; so silently ignore it if it's not for us
-   */
-  if (data == NULL)
-    goto out;
-
-  if (data->reply_handler_id == 0)
-    {
-      data->reply_handler_id = g_signal_connect (data->mount_operation,
-                                                 "reply",
-                                                 G_CALLBACK (mount_operation_reply),
-                                                 data);
-    }
-
-  g_signal_emit_by_name (data->mount_operation,
-                         "ask-question",
-                         message,
-                         choices->pdata);
-
- out:
-  g_ptr_array_free (choices, TRUE);
-}
-
-void
-g_proxy_volume_handle_mount_op_aborted (GProxyVolume        *volume,
-                                        DBusMessageIter     *iter)
-{
-  const char *mount_op_id;
-  DBusOp *data;
-
-  dbus_message_iter_get_basic (iter, &mount_op_id);
-  dbus_message_iter_next (iter);
-
-  data = g_hash_table_lookup (volume->hash_mount_op_id_to_data, mount_op_id);
-
-  /* since eavesdropping is enabled on the session bus we get this signal even if it
-   * is for another application; so silently ignore it if it's not for us
-   */
-  if (data == NULL)
-    goto out;
-
-  g_signal_emit_by_name (data->mount_operation, "aborted");
-
- out:
-  ;
-}
-
 static gboolean
 g_proxy_volume_mount_finish (GVolume        *volume,
                              GAsyncResult  *result,
@@ -1201,6 +975,8 @@ g_proxy_volume_volume_iface_init (GVolumeIface *iface)
   iface->mount_finish = g_proxy_volume_mount_finish;
   iface->eject = g_proxy_volume_eject;
   iface->eject_finish = g_proxy_volume_eject_finish;
+  iface->eject_with_operation = g_proxy_volume_eject_with_operation;
+  iface->eject_with_operation_finish = g_proxy_volume_eject_with_operation_finish;
   iface->get_identifier = g_proxy_volume_get_identifier;
   iface->enumerate_identifiers = g_proxy_volume_enumerate_identifiers;
   iface->get_activation_root = g_proxy_volume_get_activation_root;
